"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Services = require("./services");
const _1 = require("./");
const SingleProcessDiscovery_1 = require("./discovery/SingleProcessDiscovery");
const Realm = require("realm");
const _ = require("lodash");
const Logger_1 = require("./shared/Logger");
const autokeygen_1 = require("./autokeygen");
const Token_1 = require("./shared/Token");
const SyncedRealm_1 = require("./shared/SyncedRealm");
const errors = require("./errors");
const Express = require("express");
const http = require("http");
const bodyParser = require("body-parser");
const morgan = require("morgan");
const fs = require("fs-extra");
const path = require("path");
const url = require("url");
const URI = require("urijs");
const pathMatcher_1 = require("./shared/pathMatcher");
class ServerConfiguration {
}
exports.ServerConfiguration = ServerConfiguration;
const realmFeatureToken = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJTeW5jIjp0cnVlLCJpYXQiOjE0OTI2MTEzMjl9.ixnaKuxOc3AUF0-h4g5Xy9OXIg1caEEdjjehz-2kweVZcBZvmmZ59oCsdJsDCemzQnGDqcfGLDsLa5S3RseNFUeYBsDHzdg3wyYRvKuEb9Va1qwjtfIearnU0n8-X5c_M8gUjmv6f22SLVjHptRFxudRmm5ej4AP1v2eyU_hCkQgTAk9twva4xDrdJX9VKGeA2QWwetbk2BnTIuW_rzJcTAD5eAw3fyKQXbxvEte7RqStYcP7gS02uiLxWJeZ_liIoO-SOq-ySwInb8NlhMJN8jWHp-cDoRGVHwTIhyK6IkKWkkymnhDAV4BitYbaK1pY05-5IOF4v8OXXVRmZav4A";
class Server {
    constructor(config = {}) {
        this.version = require('../package.json').version;
        this.syncedRealms = {};
        this.config = config;
        if (!this.config.hostname) {
            this.config.hostname = '127.0.0.1';
        }
        if (!this.config.port) {
            this.config.port = 9080;
        }
        if (this.config.autoKeyGen === undefined) {
            this.config.autoKeyGen = true;
        }
        if (!this.config.privateKeyPath) {
            this.config.privateKeyPath = path.join(this.config.dataPath, "keys", "auth.key");
        }
        if (!this.config.publicKeyPath) {
            this.config.publicKeyPath = path.join(this.config.dataPath, "keys", "auth.pub");
        }
        this.services = [];
        this.expressApp = Express();
        this.expressApp.use(bodyParser.json());
        this.httpServer = http.createServer(this.expressApp);
        this.discovery = this.config.discovery || new SingleProcessDiscovery_1.SingleProcessDiscovery();
        this.logger = this.config.logger || new Logger_1.FileConsole(this.config.logPath, this.config.logLevel);
        this.expressApp.use(morgan("short", { "stream": { write: (msg) => { this.logger.info(msg.trim()); } } }));
        this.httpServer.on('upgrade', this.upgradeHandler.bind(this));
        Realm.Sync.setFeatureToken(process.env.ROS_FEATURE_TOKEN || realmFeatureToken);
        Realm.Sync.setSyncLogger((level, message) => {
            const logLevel = Logger_1.SyncLogLevel[level - 1];
            this.logger[logLevel](`sync-client: ${message}`);
        });
    }
    get started() {
        return this._started;
    }
    useMiddleware(...args) {
        this.expressApp.use(args);
        return this;
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._started) {
                throw new Error('The server has already been started.');
            }
            this.logger.info(`Realm Object Server version ${this.version} is starting`);
            let autoKeysWereGenerated = false;
            if (this.config.autoKeyGen) {
                autoKeysWereGenerated = yield autokeygen_1.autoGenerateKeys(path.join(this.config.dataPath, "keys"));
            }
            this.privateKey = yield fs.readFile(this.config.privateKeyPath, 'utf-8');
            this.publicKey = yield fs.readFile(this.config.publicKeyPath, 'utf-8');
            yield this.injectLogger();
            this.addServiceEndpoints();
            yield _1.Promisify(this.httpServer.listen.bind(this.httpServer), this.config.port, this.config.hostname);
            yield this.startServices();
            yield this.registerServices();
            this._started = true;
            const address = this.httpServer.address();
            for (let service of this.services) {
                const serverStartedFunctionName = Reflect.get(service, 'serverStartedFunctionName');
                if (serverStartedFunctionName) {
                    yield Promise.resolve(service[serverStartedFunctionName](this));
                }
            }
            this.logger.info(`Realm Object Server has started and is listening on ${address.address}:${address.port}`);
        });
    }
    get address() {
        if (this.httpServer) {
            const address = this.httpServer.address();
            return `${address.address}:${address.port}`;
        }
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.deregisterServices();
            yield this.stopServices();
            yield _1.Promisify(this.httpServer.close.bind(this.httpServer));
            for (const remotePath in this.syncedRealms) {
                if (this.syncedRealms[remotePath]) {
                    this.logger.debug(`closing synced realm: '${remotePath}'`);
                    this.syncedRealms[remotePath].close();
                    delete this.syncedRealms[remotePath];
                }
            }
            this._started = false;
            this.logger.info("Realm Object Server has stopped");
            if (!this.config.logger) {
                this.logger.close();
            }
        });
    }
    addServices(...services) {
        services.forEach(s => this.addService(s));
        return this;
    }
    addService(service, config) {
        if (this._started) {
            throw new Error('Services can only be added before starting the server.');
        }
        let instance = typeof (service) === 'string' ?
            new Services[service](config) : service;
        this.services.push(instance);
        return this;
    }
    removeService(service) {
        if (this._started) {
            throw new Error('Services can only be added before starting the server.');
        }
        if (typeof service === 'string') {
            for (let instance of this.services) {
                let serviceName = Reflect.get(instance.constructor, 'serviceName');
                if (serviceName === service) {
                    this.services.splice(this.services.indexOf(instance), 1);
                }
            }
        }
        else {
            const index = this.services.indexOf(service);
            if (index > -1) {
                this.services.splice(index, 1);
            }
        }
        return this;
    }
    getSyncedRealm(def) {
        if (!this.syncedRealms[def.remotePath]) {
            const localPath = path.join(this.config.dataPath, "realms", `${def.remotePath}.realm`);
            this.logger.debug(`opening synced realm: '${localPath}'`);
            this.syncedRealms[def.remotePath] = new SyncedRealm_1.SyncedRealm(this.discovery, {
                path: localPath,
                schema: def.schema,
                sync: {
                    url: `realm://default${def.remotePath}`,
                    user: Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), `realm://default${def.remotePath}`)
                }
            });
        }
        return this.syncedRealms[def.remotePath];
    }
    upgradeHandler(req, socket, head) {
        return __awaiter(this, void 0, void 0, function* () {
            const requestPath = url.parse(req.url).path;
            let numberOfMatches = 0;
            for (let service of this.services) {
                let baseRoute = Reflect.get(service.constructor, 'baseRoute');
                if (!baseRoute) {
                    continue;
                }
                let upgradeRoutes = Reflect.get(service, 'upgradeRoutes') || [];
                if (upgradeRoutes.length === 0) {
                    continue;
                }
                for (let upgradeRoute of upgradeRoutes) {
                    let concatenatedServicePath = baseRoute;
                    if (upgradeRoute.path !== '/') {
                        concatenatedServicePath = URI.joinPaths(baseRoute, upgradeRoute.path).toString();
                    }
                    const params = pathMatcher_1.pathMatcher(concatenatedServicePath, requestPath);
                    if (params === false) {
                        continue;
                    }
                    req.params = params;
                    numberOfMatches = numberOfMatches + 1;
                    let parameterArguments = Reflect.get(service, 'parameterArguments') || [];
                    parameterArguments = parameterArguments.filter(p => p.functionName === upgradeRoute.functionName);
                    const handleError = (err) => {
                        this.logger.error(err);
                        const problem = this.processException(err);
                        socket.write(`HTTP/1.1 ${problem.status}\r\nUpgrade Failed\r\n\r\n`);
                        socket.end();
                    };
                    try {
                        if (parameterArguments.length > 0) {
                            throw new Error('Decorator Annotations for @Uprade are not yet implemented');
                        }
                        const promise = service[upgradeRoute.functionName](req, socket, head);
                        if (promise instanceof Promise) {
                            yield promise.catch(handleError);
                        }
                    }
                    catch (err) {
                        handleError(err);
                    }
                }
            }
            if (numberOfMatches === 0) {
                socket.write('HTTP/1.1 404 Not Found\r\nUpgrade Failed\r\n\r\n');
                socket.end();
            }
        });
    }
    injectLogger() {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = this.services.map((service) => {
                const promise = Promise.resolve();
                let functionName = Reflect.get(service, 'unmuteFunctionName');
                if (functionName && service[functionName]) {
                    promise.then(service[functionName](this.logger));
                }
                functionName = Reflect.get(service, 'unmuteContextualFunctionName');
                if (functionName && service[functionName]) {
                    const serviceName = Reflect.get(service.constructor, 'serviceName') || service.constructor.name;
                    promise.then(service[functionName](this.logger.withContext({ service: serviceName })));
                }
                return promise;
            });
            yield Promise.all(promises);
        });
    }
    addServiceEndpoints() {
        this.expressApp.use(this.authenticationMiddleware.bind(this));
        for (const service of this.services) {
            let routes = Reflect.get(service, 'routes');
            if (!routes) {
                continue;
            }
            let baseRoute = Reflect.get(service.constructor, 'baseRoute');
            if (!baseRoute) {
                throw new Error(`${service} does not have a base route`);
            }
            const router = Express.Router();
            let serviceStaticRoutes = Reflect.get(service, 'serveStaticRoutes') || [];
            for (let serviceStaticRoute of serviceStaticRoutes) {
                router.use(serviceStaticRoute.path, Express.static(serviceStaticRoute.path));
            }
            for (let route of routes) {
                let middlewaresBefore = Reflect.get(service[route.functionName], 'middlewaresBefore');
                if (middlewaresBefore) {
                    router.use(route.path, middlewaresBefore);
                }
                let handler = (req, res, next) => __awaiter(this, void 0, void 0, function* () {
                    let functionName = route.functionName;
                    let parameterArgumentMetaData = Reflect.get(service, 'parameterArguments') || [];
                    parameterArgumentMetaData = parameterArgumentMetaData.filter(x => x.functionName === functionName);
                    try {
                        if (parameterArgumentMetaData.length > 0) {
                            let args = [];
                            for (let p of parameterArgumentMetaData) {
                                switch (p.type) {
                                    case 'body':
                                        args[p.argumentIndex] = p.keyPath ? _.get(req.body, p.keyPath) : req.body;
                                        break;
                                    case 'query':
                                        args[p.argumentIndex] = p.keyPath ? _.get(req.query, p.keyPath) : req.query;
                                        break;
                                    case 'headers':
                                        args[p.argumentIndex] = p.keyPath ? req.headers[p.keyPath.toLowerCase()] : req.headers;
                                        break;
                                    case 'params':
                                        args[p.argumentIndex] = p.keyPath ? _.get(req.params, p.keyPath) : req.params;
                                        break;
                                    case 'request':
                                        args[p.argumentIndex] = req;
                                        break;
                                    case 'response':
                                        args[p.argumentIndex] = res;
                                        break;
                                    case 'next':
                                        args[p.argumentIndex] = next;
                                        break;
                                }
                            }
                            let result = yield Promise.resolve(service[functionName](...args));
                            if (result) {
                                if (result.toJSON) {
                                    res.json(result.toJSON());
                                }
                                else {
                                    res.json(result);
                                }
                            }
                        }
                        else {
                            let result = yield Promise.resolve(service[functionName](req, res, next));
                            if (result) {
                                res.json(result);
                            }
                        }
                    }
                    catch (err) {
                        next(err);
                    }
                });
                router[route.httpMethod](route.path, handler);
            }
            this.expressApp.use(baseRoute, router);
        }
        this.expressApp.use(this.errorHandler.bind(this));
    }
    startServices() {
        return __awaiter(this, void 0, void 0, function* () {
            for (let s of this.services) {
                const startFunctionName = Reflect.get(s, 'startFunctionName');
                if (startFunctionName && s[startFunctionName]) {
                    yield Promise.resolve(s[startFunctionName](this));
                }
            }
        });
    }
    stopServices() {
        return __awaiter(this, void 0, void 0, function* () {
            for (let s of this.services) {
                const stopFunctionName = Reflect.get(s, 'stopFunctionName');
                if (stopFunctionName) {
                    yield Promise.resolve(s[stopFunctionName]());
                }
            }
        });
    }
    registerServices() {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = this.services.map((s) => {
                let address;
                const addressFunctionName = Reflect.get(s, 'addressFunctionName');
                if (addressFunctionName) {
                    address = s[addressFunctionName]();
                }
                else {
                    address = this.httpServer.address();
                }
                return this.discovery.registerService(s, address.address, address.port);
            });
            yield Promise.all(promises);
        });
    }
    deregisterServices() {
        return __awaiter(this, void 0, void 0, function* () {
            for (let service of this.services) {
                if (!service) {
                    yield this.discovery.deregisterService(service);
                }
            }
        });
    }
    authenticationMiddleware(req, res, next) {
        try {
            const authorization = req.headers['authorization'];
            if (authorization) {
                const tokenData = authorization;
                try {
                    req.authToken = Token_1.Token.parse(tokenData, this.publicKey);
                }
                catch (err) {
                    next(new errors.realm.InvalidCredentials());
                }
            }
        }
        catch (err) {
        }
        next();
    }
    processException(error) {
        if (error instanceof errors.JSONError) {
            return error;
        }
        else {
            const status = error.status || error.statusCode;
            if (status) {
                return new errors.JSONError({
                    status: status,
                    stack: error.stack,
                    title: error.message,
                });
            }
            return new errors.JSONError({
                status: 500,
                stack: error.stack,
                title: 'Internal Server Error',
            });
        }
    }
    errorHandler(error, req, res, next) {
        if (res.headersSent) {
            next(error);
            return;
        }
        const jsonError = this.processException(error);
        res.setHeader('Content-Type', 'application/json');
        res.status(jsonError.status).json(jsonError.toJSON());
    }
}
exports.Server = Server;
//# sourceMappingURL=Server.js.map