"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const auth = require("../auth");
const TestServer_1 = require("../TestServer");
const services = require("../services");
const chai_1 = require("chai");
const chai = require("chai");
const Realm = require("realm");
const uuid = require("uuid");
const path = require("path");
const faker = require("faker");
chai.use(require("chai-as-promised"));
describe('Realm-JS integration', function () {
    const ProductSchema = {
        name: 'Product',
        primaryKey: 'productId',
        properties: {
            productId: { type: 'int', optional: false },
            name: { type: 'string', optional: false },
            price: { type: 'float', optional: false }
        }
    };
    let server;
    before(() => __awaiter(this, void 0, void 0, function* () {
        server = new TestServer_1.TestServer();
        const authService = new services.AuthService();
        authService.addProvider(new auth.PasswordAuthProvider({ autoCreateAdminUser: true, saltWorkIterations: 1 }));
        server.addService(authService);
        server.addService(new services.RealmDirectoryService());
        server.addService(new services.SyncProxyService());
        server.addService(new services.SyncService());
        yield server.start();
    }));
    after(() => __awaiter(this, void 0, void 0, function* () {
        if (server) {
            yield server.shutdown();
        }
    }));
    it('should be able to register to realm-js', function () {
        return __awaiter(this, void 0, void 0, function* () {
            yield chai_1.assert.isFulfilled(Realm.Sync.User.register(server.url, 'sampleuser', 'samplepassword'));
        });
    });
    it('should be able to login to realm-js', function () {
        return __awaiter(this, void 0, void 0, function* () {
            const registeredUser = yield Realm.Sync.User.register(server.url, 'newuser', 'samplepassword');
            registeredUser.logout();
            yield chai_1.assert.isFulfilled(Realm.Sync.User.login(server.url, 'newuser', 'samplepassword'));
        });
    });
    it('should be able to login to realm-js with an empty string password', function () {
        return __awaiter(this, void 0, void 0, function* () {
            yield chai_1.assert.isFulfilled(Realm.Sync.User.login(server.url, 'realm-admin', ''));
        });
    });
    it('should create error when trying to login with an unknown username', function () {
        return __awaiter(this, void 0, void 0, function* () {
            yield chai_1.assert.isRejected(Realm.Sync.User.login(server.url, 'sameuser', 'samplepassword'), 'The provided credentials are invalid or the user does not exist.');
        });
    });
    it('should create error when trying to register an existing username', function () {
        return __awaiter(this, void 0, void 0, function* () {
            const registeredUser = yield Realm.Sync.User.register(server.url, 'sameuser', 'samplepassword');
            registeredUser.logout();
            yield chai_1.assert.isRejected(Realm.Sync.User.register(server.url, 'sameuser', 'samplepassword'), 'The provided credentials are invalid or the user does not exist.');
        });
    });
    function registerAndGetRealm(username, password, localPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const user = yield Realm.Sync.User.register(server.url, username, password);
            return new Realm({
                path: localPath,
                sync: {
                    url: `realm://${server.address}/~/products`,
                    user,
                },
                schema: [ProductSchema]
            });
        });
    }
    function loginAndGetRealm(username, password, localPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const user = yield Realm.Sync.User.login(server.url, username, password);
            return new Realm({
                path: localPath,
                sync: {
                    url: `realm://${server.address}/~/products`,
                    user,
                },
                schema: [ProductSchema]
            });
        });
    }
    it('should be able to write to a realm', () => __awaiter(this, void 0, void 0, function* () {
        const USERNAME = uuid.v4();
        const localPath1 = path.join(server.config.dataPath, "local1.realm");
        const realm1 = yield registerAndGetRealm(USERNAME, 'samplepassword', localPath1);
        const localPath2 = path.join(server.config.dataPath, "local2.realm");
        const realm2 = yield loginAndGetRealm(USERNAME, 'samplepassword', localPath2);
        return new Promise((resolve, reject) => {
            let obj1;
            realm2.addListener('change', (realm, event) => {
                const obj2 = realm2.objectForPrimaryKey('Product', 0);
                chai_1.assert.equal(obj1.name, obj2.name);
                chai_1.assert.equal(obj1.price, obj2.price);
                resolve();
            });
            realm1.write(() => {
                obj1 = realm1.create('Product', {
                    productId: 0,
                    name: faker.commerce.productName(),
                    price: faker.random.number({ min: 12, max: 7000 })
                }, true);
            });
        });
    }));
});
//# sourceMappingURL=realmjs-integration-tests.spec.js.map